import json
import traceback

import yaml

from definitions import DEBUG, LOG

from .enums import ContentType, Status
from .request import Request


class Response():

    def __init__(self, status:Status=Status.OK):
        self.response_body = None
        self.status = status
        self.headers = dict()
        self.set_content_type(ContentType.PLAIN)

    def body(self, response_body:str):
        self.response_body = response_body
        return self

    def html(self, html:str):
        self.set_content_type(ContentType.HTML)
        return self.body(json.dumps(html))

    def yaml(self, response:dict):
        self.set_content_type(ContentType.YAML)
        return self.body(yaml.dump(response))

    def json(self, response:dict):
        self.set_content_type(ContentType.JSON)
        return self.body(json.dumps(response))

    def get_status_code(self):
        return self.status.value[1]

    def set_content_type(self, content_type:ContentType):
        self.header('Content-Type', content_type.value)

    def header(self, header_name, header_value):
        self.headers[header_name] = header_value
        return self

    def not_found(self, request:Request):
        self.status = Status.NOT_FOUND
        return self.__set_response(request, 'Resource not found')

    def forbidden(self, request:Request):
        self.status = Status.FORBIDDEN
        return self.__set_response(request, 'Fobidden')

    def server_error(self, request:Request, exc_info=None):
        self.status = Status.SERVER_ERROR
        data = {}
        if exc_info:
            if DEBUG:
                data['exception'] = traceback.format_exception(*exc_info)
            else:
                LOG.error(self.status.value, exc_info=exc_info)
        return self.__set_response(request, 'Internal server error', data)

    def __set_response(self, request:Request, message, data:dict={}):
        if request.get_header('ACCEPT') == 'application/json':
            self.json({
                'status': self.status.value[0],
                'message': message,
                **data
            })
        else:
            self.body(message + "\n" + str(data) if data else message)
        return self



from enum import Enum


class Status(Enum):
    OK = [200, '200 Ok']
    CREATED = [201, '201 Created']
    ACCEPTED = [202, '202 Accepted']
    NO_CONTENT = [204, '204 No content']
    MOVED_PERMANENTLY = [301, '301 Moved permanently']
    FOUND = [302, '302 Found']
    BAD_REQUEST = [400, '400 Bad request']
    UNAUTHORIZED = [401, '401 Unauthorized']
    FORBIDDEN = [403, '403 Forbidden']
    NOT_FOUND = [404, '404 Not found']
    METHOD_NOT_ALLOWED = [405, '405 Method not allowed']
    SERVER_ERROR = [500, '500 Server error']


class ContentType(Enum):
    PLAIN = 'text/plain'
    HTML = 'text/html'
    JSON = 'application/json'
    YAML = 'application/x-yaml'




import ast
import importlib
import inspect

import definitions


class OpenApi:

    OPEN_API_VERSION = '3.0.3'

    def generate(self, kaa):
        elements = self.__get_elements(kaa)
        return {
            'openapi': self.OPEN_API_VERSION,
            'info': {
                'title': definitions.NAME,
                'version': definitions.VERSION
            },
            'paths': self.__get_paths(elements),
            # 'components': {
            #     'schemas': {
            #         'Default': {}
            #     }
            # }
        }

    def __get_elements(self, kaa):
        elements = []
        for module_name in kaa.resources:
            for class_name in kaa.resources[module_name]:
                elements.append(self.__fn(module_name, class_name))
        return elements

    def __get_paths(self, elements):
        paths = {}
        for element in elements:
            for operation_id in element:
                resource = element[operation_id]
                parameters = []

                if 'path_params' in resource:
                    parameters += self.__path_params(resource['path_params'])
                if 'query_params' in resource:
                    parameters += self.__query_params(resource['query_params'])

                method = resource['method'].lower()
                paths[resource['uri']] = {
                    method: {
                        'operationId': operation_id,
                        'parameters': parameters,
                        'description': self.__get_val(lambda v: v, resource, 'description', ''),
                        'responses': {
                            '200': {
                                'description': 'Response description',
                                'content': {'application/json':{}},
                                # '$ref': '#/components/schemas/Default'
                            }
                        }
                    }
                }
        return paths

    def __fn(self, module_name, class_name):
        module = importlib.import_module(module_name)
        class_ = getattr(module, class_name)
        return get_decorators(class_)

    def __path_params(self, parameters):
        pm = []
        for k in parameters:
            parameter = parameters[k]
            p = {
                'name': k,
                'description': self.__get_val(lambda v: v, parameter, 'description', ''),
                'in': 'path',
                'schema': {
                    'type': self.__get_val(self.__get_type, parameter, 'type', 'string')
                },
                'required': True,
            }
            pm.append(p)
        return pm

    def __get_val(self, fn, item, key, default):
        return fn(item[key]) if key in item else default

    def __query_params(self, parameters):
        pm = []
        for k in parameters:
            parameter = parameters[k]
            p = {
                'name': k,
                'description': self.__get_val(lambda v: v, parameter, 'description', ''),
                'in': 'query',
                'schema': {
                    'type': self.__get_val(self.__get_type, parameter, 'type', 'string')
                },
                'required': self.__get_val(lambda v: v, parameter, 'required', False),
            }
            pm.append(p)
        return pm

    def __get_type(self, param_type):
        if param_type == 'int':
            return 'integer'
        elif param_type == 'float':
            return 'number'
        return 'string'


def get_decorators(cls):
    AVAILABLE_METHODS = ['GET', 'PUT', 'POST', 'DELETE']
    PATH = 'PATH'
    AUTH = 'AUTH'
    target = cls
    decorators = {}

    def visit_fn(node):
        if node.name[:2] == "__":
            return
        decorators[node.name] = {}
        for n in node.decorator_list:
            if isinstance(n, ast.Call):
                name = n.func.attr if isinstance(n.func, ast.Attribute) else n.func.id
                if name == PATH:
                    decorators[node.name].update(parse_path(n))
                elif name == AUTH:
                    decorators[node.name]['authorization'] = True
            else:
                name = n.attr if isinstance(n, ast.Attribute) else n.id
                if name in AVAILABLE_METHODS:
                    decorators[node.name]['method'] = name

    node_iter = ast.NodeVisitor()
    node_iter.visit_FunctionDef = visit_fn
    node_iter.visit(ast.parse(inspect.getsource(target)))
    return decorators


def parse_path(node):
    result = {}

    def _parse(node):
        if isinstance(node, ast.AST):
            fields = [(a, b) for a, b in ast.iter_fields(node)]
            for field in fields:
                if field[0] == 'args':
                    result.update(_parse_args(field[1]))
                elif field[0] == 'keywords':
                    result.update(_parse_kw(field[1]))
                elif isinstance(field[1], ast.AST):
                    _parse(field[1])

    def _parse_args(node):
        if not isinstance(node, list):
            return {}
        childs = {}
        count = 0
        for child in node:
            count += 1
            if count == 1:
                childs['uri'] = _get_iter_fields(child)[0]
            if count == 2:
                childs['query_params'] = _parse_dict(child)
        return childs

    def _parse_kw(node):
        childs = {}
        for child in node:
            fields = _get_iter_fields(child)
            childs[fields[0]] = _parse_dict(fields[1])
        return childs

    def _parse_dict(node):
        params = []
        values = []
        for field_key, field_value in ast.iter_fields(node):
            if field_key == 'keys':
                for k in field_value:
                    params += _get_iter_fields(k)
            elif field_key == 'values':
                for k in field_value:
                    fields = _get_iter_fields(k)
                    if len(fields) == 1:
                        values.append(fields[0])
                    else:
                        values.append(_parse_dict(k))
            else:
                return field_value
        idx = 0
        result = {}
        for param in params:
            result[param] = values[idx]
            idx += 1
        return result

    def _get_iter_fields(node):
        return [v for k, v in ast.iter_fields(node)]

    _parse(node)
    return result



import importlib
from definitions import SERVER
from kaa import KaaServer


class Server:

    def get_server(self) -> KaaServer:
        spl = SERVER.split('.')
        class_name = spl[-1]
        module_name = '.'.join(spl[:-1])
        module = importlib.import_module(module_name)
        class_ = getattr(module, class_name)
        return class_()

    def serve(self, env, start_response):
        server:KaaServer = self.get_server()
        return server.serve(env, start_response)



import json
import sys
from wsgiref.simple_server import make_server

from . import NAME, VERSION, Kaa, KaaServer
from .openapi import OpenApi
from .server import Server


class Cli():

    def __init__(self):
        self.host = '127.0.0.1'
        self.port = 8086
        self.argv = sys.argv[:]

    def execute(self):
        try:
            subcommand = self.argv[1]
        except IndexError:
            subcommand = 'help'

        if subcommand == 'version':
            msg = self.__get_version
        elif subcommand == 'help':
            msg = self.__get_help()
        elif subcommand == 'serve':
            self.__serve()
            return
        elif subcommand == 'openapi':
            server:KaaServer = Server().get_server()
            kaa:Kaa = server.get_kaa({
                'REQUEST_METHOD': '',
                'PATH_INFO': '',
                'REMOTE_ADDR': '',
                'QUERY_STRING': ''
            }, None)

            msg = json.dumps(OpenApi().generate(kaa))
        else:
            msg = 'Invalid command. Try help'

        sys.stdout.write(msg + '\n')

    def __get_name(self):
        return NAME

    def __get_version(self):
        return VERSION

    def __get_help(self):
        commands = [
            ('version', 'Returns Kaa version'),
            ('serve', 'Starts a server for development'),
            ('openapi', 'Generates openapi json')
        ]
        return '\n'.join(['{}\t\t{}'.format(*cmd) for cmd in commands])

    def __serve(self):
        self.__set_host_port()
        sys.stdout.write('{} version {}\n'.format(self.__get_name(), self.__get_version()))
        sys.stdout.write('Server started at {}:{}\n\n'.format(self.host, self.port))
        server:KaaServer = Server().get_server()
        make_server(
            host=self.host,
            port=int(self.port),
            app=lambda env, start_response: server.get_kaa(env, start_response).serve()
        ).serve_forever()

    def __set_host_port(self):
        try:
            porthost = self.argv[2].split(':')
            if len(porthost) == 1:
                self.port = porthost[0]
            elif len(porthost) == 2:
                self.host = porthost[0]
                self.port = porthost[1]
            else:
                sys.stdout.write('Invalid host:port' + '\n')
                sys.exit(1)
        except IndexError:
            pass



import re

from .authorization import Authorization
from .exceptions import InvalidParamError
from .request import Request


def AUTH(auth: Authorization):
    def decorator_path(func):
        def wrapper(self, **kwargs):
            if auth.authorize(self.request):
                return func(self, **kwargs)
            return auth.forbidden(self.request)
        return wrapper
    return decorator_path


def GET(func):
    def wrapper(self, **kwargs):
        if self.request.method == 'GET':
            return func(self, **kwargs)
    return wrapper


def POST(func):
    def wrapper(self, **kwargs):
        if self.request.method == 'POST':
            return func(self, **kwargs)
    return wrapper


def PUT(func):
    def wrapper(self, **kwargs):
        if self.request.method == 'PUT':
            return func(self, **kwargs)
    return wrapper


def DELETE(func):
    def wrapper(self, **kwargs):
        if self.request.method == 'DELETE':
            return func(self, **kwargs)
    return wrapper


def PATH(uri, query_params:dict={}, **kwargs):
    def decorator_path(func):
        def wrapper(self):
            request_path = self.request.path.strip("/")
            defined_path = uri.strip("/")

            q_params = QueryParams(self.request)
            if request_path == defined_path:
                params = q_params.get_params(query_params)
                return func(self, **params)

            split_request_path = request_path.split('/')
            split_defined_path = defined_path.split('/')

            if (len(split_request_path) != len(split_defined_path)):
                return

            regexp = r'^\{([a-z_][a-zA-Z0-9_]+)(:[^}]+){0,1}\}$'
            arguments = dict()
            for i in range(len(split_defined_path)):
                if split_defined_path[i] == split_request_path[i]:
                    continue
                m = re.search(regexp, split_defined_path[i])

                if m is None:
                    return

                if m.group(2):
                    pattern = re.compile(m.group(2)[1:])
                    n = re.search(pattern, split_request_path[i])
                    if (n is None):
                        return
                arguments[m.group(1)] = split_request_path[i]
                arguments.update(q_params.get_params(query_params))
            return func(self, **arguments)
        return wrapper
    return decorator_path


class QueryParams:

    def __init__(self, request:Request):
        self.request = request

    def get_params(self, defined_params:dict={}):
        params = {}
        for k in defined_params:
            params[k] = self.__get_item(k, defined_params[k])
        return params

    def __get_item(self, defined_key, defined_value):
        q_value = self.request.get_query_param(defined_key)
        if q_value:
            if 'type' not in defined_value:
                return q_value
            if defined_value['type'] == 'int':
                fn = lambda v: int(v)
            elif defined_value['type'] == 'float':
                fn = lambda v: float(v)
            else:
                return q_value
            return self.__get_number(defined_key, q_value, fn)

        if 'required' in defined_value and defined_value['required']:
            raise InvalidParamError("Param {} is required".format(defined_key))

        if 'default' in defined_value:
            return defined_value['default']

    def __get_number(self, param, value, fn):
        try:
            return fn(value)
        except ValueError:
            raise InvalidParamError("Param {} is not a number".format(param))



from .request import Request
from .response import Response


class RequestFilter():

    def filter(self, request:Request):
        raise NotImplementedError


class ResponseFilter():

    def filter(self, request:Request, response:Response):
        raise NotImplementedError



from .request import Request
from .response import Response


class Authorization:

    def authorize(self, request:Request) -> bool:
        raise NotImplementedError

    def forbidden(self, request:Request):
        return Response().forbidden(request)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kaa-rest-server",
    version="0.1.3",
    author="Francesc d'Assís Requesens i Roca",
    author_email="francesc.requesens@gmail.com",
    description="A very simple REST server",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/elpeix/kaa",
    packages=setuptools.find_packages(
        exclude=[
            "*.tests", "*.tests.*", "tests.*", "tests",
            "example.*","example"
        ]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=['pyyaml']
)



import sys
import traceback

from .enums import Status
from .response import Response


class KaaError(Exception):

    def __init__(self, msg, status:Status = Status.SERVER_ERROR, err=None):
        self.msg = msg
        self.status = status
        self.err = err

    def response(self):
        data = {
            "status": self.status.value[0],
            "message": self.msg
        }
        if (self.err):
            exc_info = sys.exc_info()
            data['exception'] = traceback.format_exception(*exc_info)
        return Response(self.status).json(data)

    def __str__(self):
        if self.err is None:
            return "{status} - {msg}".format(
                status=self.status.value[1],
                msg=self.msg
            )
        return "{status} - {msg} - {err}".format(
            status=self.status.value[1],
            msg=self.msg,
            err=repr(self.err))


class NotFoundError(KaaError):
    def __init__(self, msg, err=None):
        super().__init__(msg, Status.NOT_FOUND, err)


class ResourceNotFoundError(KaaError):
    def __init__(self):
        super().__init__('Resource not found', Status.NOT_FOUND)


class BadRequestError(KaaError):
    def __init__(self, msg, err=None):
        super().__init__(msg, Status.BAD_REQUEST, err)


class UnathorizedError(KaaError):
    def __init__(self, msg, err=None):
        super().__init__(msg, Status.UNAUTHORIZED, err)


class ForbiddenError(KaaError):
    def __init__(self, msg, err=None):
        super().__init__(msg, Status.FORBIDDEN, err)


class MethodNotAllowedError(KaaError):
    def __init__(self, msg, err=None):
        super().__init__(msg, Status.METHOD_NOT_ALLOWED, err)


class InvalidParamError(BadRequestError):
    pass



from .authorization import Authorization
from .decorators import AUTH, GET, PATH, POST, PUT, DELETE
from .enums import ContentType, Status
from .exceptions import *
from .filters import RequestFilter, ResponseFilter
from .kaa import Kaa
from .request import Request
from .resources import Resources
from .response import Response

NAME = 'KAA'
VERSION = '0.1'


class KaaServer:

    def get_kaa(self, env, start_response):
        pass

    def serve(self, env, start_response):
        return self.get_kaa(env, start_response).serve()

    def generate_openapi(self):
        pass



class Request():

    def __init__(self, env):
        self.env = env
        self.method = self.env['REQUEST_METHOD']
        self.path = self.env['PATH_INFO']
        self.remote_addr = self.env['REMOTE_ADDR']
        self.query = self.__get_query()
        self.headers = self.__get_headers()

    def __get_query(self):
        query_params = {}
        query_string = self.env['QUERY_STRING']
        if not query_string:
            return query_params
        for item in query_string.split('&'):
            values = item.split('=')
            if len(values) == 2:
                self.__set_query_value(values[0], values[1], query_params)
        return query_params

    def __set_query_value(self, key, value, query_params):
        if key in query_params:
            if type(query_params[key]) == list:
                query_params[key].append(value)
            else:
                query_params[key] = [query_params[key], value]
        else:
            query_params[key] = value

    def __get_headers(self):
        headers = {}
        for item in self.env:
            if item[:5] == 'HTTP_':
                headers[item[5:]] = self.env[item]
        return headers

    def get_header(self, key):
        if key in self.headers:
            return self.headers[key]
        return None

    def get_query_param(self, key):
        if key in self.query:
            return self.query[key]
        return None

    def get_dict(self):
        return {
            'method': self.method,
            'path': self.path,
            'remoteAddr': self.remote_addr,
            'query': self.query,
            'headers': self.headers
        }



import importlib
import sys

import kaa
from definitions import ENABLE_CORS

from .enums import Status
from .exceptions import KaaError, ResourceNotFoundError
from .filters import RequestFilter, ResponseFilter
from .openapi import OpenApi
from .request import Request
from .resources import Resources
from .response import Response


class Kaa():

    def __init__(self, env, start_response):
        self.start_response = start_response
        self.request = Request(env)
        self.resources = {}
        self.request_filters = {}
        self.response_filters = {}

    def register_resources(self, module:str, class_name:str):
        self.__register(self.resources, module, class_name)

    def register_filter_request(self, module:str, class_name):
        self.__register(self.request_filters, module, class_name)

    def register_filter_response(self, module:str, class_name):
        self.__register(self.response_filters, module, class_name)

    def __register(self, element, module, class_name):
        if module not in element:
            element[module] = []
        element[module].append(class_name)

    def serve(self):
        try:
            if self.request.method == 'OPTIONS':
                return self.__act_method_options()

            if self.request.path == '/openapi':
                return self.__get_openapi()

            self.__request_filters()
            for module_name in self.resources:
                for class_name in self.resources[module_name]:
                    response:Response = self.__run_resource(module_name, class_name)
                    if response:
                        self.__response_filters(response)
                        return self.__print_response(response)
            raise ResourceNotFoundError()
        except KaaError as e:
            return self.__print_response(e.response())
        except Exception:
            return self.__print_response(Response().server_error(self.request, sys.exc_info()))

    def __get_openapi(self):
        openapi = OpenApi().generate(self)
        response = Response()
        if self.request.get_header('ACCEPT') == 'application/json':
            response.json(openapi)
        else:
            response.yaml(openapi)
        return self.__print_response(response)

    def __act_method_options(self):
        if ENABLE_CORS:
            response = Response(Status.ACCEPTED)
            self.__response_filters(response)
        else:
            response = Response(Status.METHOD_NOT_ALLOWED)
        return self.__print_response(response.body(''))

    def __request_filters(self):
        def func(instance:RequestFilter):
            method_ = getattr(instance, 'filter')
            method_(instance, self.request)
        self.__call_filters(self.request_filters, func)

    def __response_filters(self, response:Response):
        def func(instance:ResponseFilter):
            method_ = getattr(instance, 'filter')
            method_(instance, self.request, response)
        self.__call_filters(self.response_filters, func)

    def __call_filters(self, filters, func):
        for module_name in filters:
            for class_name in filters[module_name]:
                func(self.__get_class(module_name, class_name))

    def __run_resource(self, module_name, class_name) -> Response:
        class_ = self.__get_class(module_name, class_name)
        instance:Resources = class_(self.request)
        for method_name in dir(class_):
            count = 1 + len(class_name) + 2
            if method_name[:2] == "__" or method_name[:count] == "_{}__".format(class_name):
                continue
            method_ = getattr(class_, method_name)
            result = method_(instance)
            if result:
                return result

    def __get_class(self, module_name, class_name):
        module = importlib.import_module(module_name)
        return getattr(module, class_name)

    def __print_response(self, response:Response):
        headers = [(k, response.headers[k]) for k in response.headers]
        headers.append(('Server', '{}/{}'.format(kaa.NAME, kaa.VERSION)))
        self.start_response(response.get_status_code(), headers)
        return [response.response_body.encode("utf8")]
